
// ExecutePixels.java
// Andrew Davison, January 2009, ad@fivedots.coe.psu.ac.th

/* Execute the supplied PNG file as a Java application.

   First, the pixels in the PNG file are converted into a string of ASCII chars.
   using ImageText.pngToString(). The string is treated as the text 
   of a Java application, assuming that all the program is contained
   in the PNG file, and has a main() method.

   Janino's SimpleCompiler API is used to compile the Java text as a single 
   "compilation unit" (i.e. treating the text as if it were the contents
   of a single ".java" file). 
   Janino can be downloaded from http://www.janino.net/

   The compiled code is executed using reflection to start the application's
   main() method (which we assume doesn't use its args[] input argument).

   While the translation/compilation/execution phases are being carried
   out, ExecutePixels appears as a small GUI containing a progress
   text field, and progress bar. A timer is used to periodically update 
   the text field with a changing progress message string.

   Usage: java ExecutePixels <PNG file>
   e.g.
      java ExecutePixels Comparison.png
*/

import java.io.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.util.regex.*;
import java.lang.reflect.*;

import org.codehaus.janino.*;  
   // requires janino.jar, available from http://www.janino.net/



public class ExecutePixels extends JFrame
{  
  // GUI elements accessed by the timer's listener
  private JProgressBar progressBar;
  private JTextField msgJTF;
  private String progressMsg = "Starting...";

  private boolean foundError = false;   
    // set is there's an error in the translation/compilation/execution phases


  public ExecutePixels(String[] args) 
  {
    super("EP");

    buildGUI();
    setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    pack();
    setResizable(false);
    setVisible(true);

    startReporting();

    execute(args);
  }  // end of ExecutePixels()



  private void buildGUI()
  // A text field for messages, and a progress bar
  {
    Container c = getContentPane();
    c.setLayout(new BoxLayout(c, BoxLayout.Y_AXIS)); // vertical box layout

    setBackground(Color.white);

    msgJTF = new JTextField(5);
    msgJTF.setEditable(false);
    c.add(msgJTF);

    JPanel p1 = new JPanel();
    progressBar = new JProgressBar();
    progressBar.setIndeterminate(true);
    p1.add(progressBar);
    c.add(p1);
  }  // end of buildGUI()



  private void startReporting()
  // Fire a timer periodically to report on things
  {
    javax.swing.Timer timer =    // call the listener every 500ms
       new javax.swing.Timer(500, new ActionListener() {
            public void actionPerformed(ActionEvent e) 
            { if (foundError) {     
                // stop the progress bar when there's an error
                progressBar.setIndeterminate(false);
                progressBar.setValue(20);
              }
              msgJTF.setText(progressMsg);  // update the message text field
            }
          });

    // timer.setInitialDelay(500);
    timer.start();
  }  // end of startReporting()



  // -------------------- compilation & execution ---------------


  private void execute(String[] args)
  // translate image to code and execute
  {
    if (args.length < 1)
      reportError("No image supplied");
    else {
      progressMsg = "Creating Code...";
      String codeStr = ImageText.pngToString(args[0]);   // translate image --> code
      if (codeStr == null)
        reportError("No code to execute");
      else
        execCode(codeStr);
    }
  }  // end of execute()



  private void reportError(String msg)
  /* Use a dialog box to diaplay an error, and set the 
     error flag to update the GUI display. */
  { 
     foundError = true;
     progressMsg = "Error";
     JOptionPane.showMessageDialog(null, msg,
        "ExecutePixels Error", JOptionPane.WARNING_MESSAGE);
     System.exit(1);
  }  // end of reportError()



  private void execCode(String codeStr)
  /* Compile the text in CodeStr using Janino's SimpleCompiler API,
     load the class, and call it's main() method with no input 
     arguments.
  */
  {
    String className = getClassName(codeStr);   // public class name used in codeStr

    try {
      progressMsg = "Compiling " + className + "...";
      Class.forName( "org.codehaus.janino.SimpleCompiler" );  
                   // check if Janino's SimpleCompiler is present

      // compile the code in codeStr
      SimpleCompiler compiler = new SimpleCompiler();
      compiler.cook( new StringReader(codeStr) );

      // get the loaded class
      Class<?> cl = compiler.getClassLoader().loadClass(className);

      // get a reference to the class' main(String[]) method
      progressMsg = "Running " + className + "...";
      Method mainMeth = cl.getMethod("main", new Class[] { String[].class });

      String[] args = new String[] { };    // no inputs supplied to main()
      mainMeth.invoke(null, new Object[] { args} );   // call main()
    }
    catch(Exception ex)
    {  
      if (ex instanceof InvocationTargetException) {
        Throwable t = ((InvocationTargetException) ex).getTargetException();
        reportError(t.toString());
        // System.out.println(t);
      }
      else {
        reportError(ex.toString());
        // System.out.println(ex);  
      }
    }
  }  // end of execCode()



  private String getClassName(String codeStr)
  // return the identifier that appears after "public class" 
  {
    String className = null;

    // prepare regular expression
    String idPatternStr = "[a-zA-Z_$][\\w$]*";    
             // RE string for a Java ID
    String codePatternStr = "public\\s*class\\s*(" + idPatternStr+ ")";
             // RE string for "public class (ID)"  (...) is for grouping
    Pattern codePattern = Pattern.compile(codePatternStr);

    Matcher codeMatcher = codePattern.matcher(codeStr);  // search code

    if (codeMatcher.find() && (codeMatcher.groupCount() == 1)) {
      // found a single ID
      className = codeMatcher.group(1);
      // System.out.println("Found class name: " + className);
    }
    else {
      reportError("Could not find class name in code");
      // System.out.println("Could not find class name in code");
    }
    
    return className;
  }  // end of getClassName()



  // -------------------------------------------------------

  public static void main(String[] args)
  {  new ExecutePixels(args);  } 


}  // end of ExecutePixels class