
// ImageText.java
// Andrew Davison, January 2009, ad@fivedots.coe.psu.ac.th

/* This class contains static methods for converting 
   text to images, and images into text.

   Public "text --> image" methods:
          * boolean textToPng(String fnm);
          * BufferedImage textToImage(String fnm);
   
   textToPng() reads the ASCII text from a file (fnm), 
   and uses it to create a square-ish BufferedImage, which 
   is saved as a PNG file with the same filename + ".png".
   textToImage() returns the BufferedImage without saving it to a file.

   Public "image --> text" methods
          * boolean pngToText(String imFnm);
          * String pngToString(String imFnm);

   pngToText() converts the pixels in a PNG file (imFnm)
   into ASCII text, and stores it in a text file with the 
   same filename + ".txt". 
   pngToString() returns the text without saving it to a file.

*/

import java.awt.*;
import java.io.*;
import java.util.*;
import java.awt.image.*;
import javax.imageio.*;


public class ImageText
{
  private static final int BRIGHTNESS = 2;
       /* scale factor for increasing the brightness of the
          colours stored in the image pixels */


  // ---------------- text --> PNG/image methods -------------------

  public static boolean textToPng(String fnm)
  /* Read the ASCII text from the fnm file and create
     a square-ish BufferedImage in a PNG file with the same name. */
  {
    BufferedImage im = textToImage(fnm);    // get the image
    if (im == null)
      return false;
    else {
      String fileName = getFileName(fnm);
      if (fileName == null)
         return false;
      else 
        return writeImageToFile(fileName+".png", im);   // save the image
    }
  }  // end of textToPng()



  public static BufferedImage textToImage(String fnm)
  /* Read the ASCII text from the fnm file and return
     a square-ish BufferedImage (or null). */
  {
    ArrayList<Integer> pixColors = new ArrayList<Integer>();
    if (readCode(fnm, pixColors))
      return createImage(pixColors);
    else
      return null;
  }  // end of textToImage()



  private static boolean readCode(String fnm, ArrayList<Integer> pixColors)
  /* Read the ASCII chars from the fnm textfile, and store them as integers 
     in pixColors. Each printable 7-bit ASCII character is scaled by 
     BRIGHTNESS to increase the final image's brightness. */
  {
    int numUnPrintables = 0;     // number of unprintable chars
    try {
      FileReader fr = new FileReader(fnm);
      int ch;
      while ((ch = fr.read()) != -1) {
        if (!isPrintable(ch)) {  // if character not printable
          ch = 32;    // replace with a space char
          numUnPrintables++;
        }
        pixColors.add(ch*BRIGHTNESS);    // increase brightness
      }
      fr.close();

      if (numUnPrintables > 0) 
        System.out.println( numUnPrintables + 
                    " chars unprintable; all set to spaces");

      System.out.println("Read in: " + fnm);
    } 
    catch (IOException e) 
    { System.out.println("Error reading: " + fnm);
      return false;
    }

    return true;
  }  // end of readCode()



  private static boolean isPrintable(int ch)
  // is ch a 7-bit ASCII character that could (sensibly) be printed?
  {
    if (Character.isWhitespace(ch) && (ch < 127))  // whitespace, 7-bit
      return true;
    else if ((ch > 32) && (ch < 127))
      return true;

    return false;
  }  // end of isPrintable()



  private static BufferedImage createImage(ArrayList<Integer> pixColors)
  /* Store the pixColors integers in a BufferedImage. Three integers are 
     packed into one pixel as its R, G, and B values. The alpha byte of the
     pixel isn't used. */
  {
    Dimension imDim = calcImageDimensions(pixColors);

    // use an array to store the pixel values
    int len = imDim.width * imDim.height;
    int[] pixels = new int[len];

    int idx = 0;
    int redVal, greenVal, blueVal; 
    for (int i=0; i < len; i++) {
      redVal = pixColors.get(idx);     // get values for the RGB components
      greenVal = pixColors.get(idx+1);
      blueVal = pixColors.get(idx+2);

      // store the three values as a single pixel value
      pixels[i] = blueVal | (greenVal << 8) | (redVal << 16); 
      idx += 3;
    }

    // create a bufferedImage for the pixels array
    BufferedImage im = new BufferedImage(imDim.width, imDim.height, 
                                               BufferedImage.TYPE_INT_RGB);
    
    // fill the image with the pixels array data
    im.setRGB(0, 0, imDim.width, imDim.height, pixels, 0, imDim.width);
    
    if (im == null)
      System.out.println("No image created");

    return im;
  }  // end of createImage()



  private static Dimension calcImageDimensions(ArrayList<Integer> pixColors)
  /* Use the length of the pixColors array to calculate a desired width and
     height for the image. The two values should be roughly equal to make 
     the image square-ish. */
  {
    int imWidth = 0;   // starting values for the image's dimensions
    int imHeight = 0;

    int numColors = pixColors.size();
    /* make numColors a multiple of three, since 3 colors are needed
       for each pixel (its R, G, and B values) */
    if (numColors%3 != 0) {
      int extras = 3 - numColors%3;
      numColors += extras;  
    }

    int pixLen = numColors/3;    // the no. of pixels that will be in the image

    // calculate roughly square width and height values for the image
    imWidth = (int) Math.sqrt(pixLen);
    imHeight = pixLen/imWidth;

    /* increase the height value until width*height is greater (or equal)
       to the number of pixels in the image */
    while ((imWidth*imHeight) < pixLen)
      imHeight++;

    /* add 0's to the pixColors list as filler so the list size
       matches the image dimensions */
    int numFillers = (imWidth*imHeight*3) - pixColors.size();
    System.out.println("Added " + numFillers + " filler colors");
    for (int i=0; i < numFillers; i++)
      pixColors.add(0);

    Dimension imDim =  new Dimension(imWidth, imHeight);
    return imDim;
  }  // end of calcImageDimensions()


  // ---------------- PNG --> text file/string methods ----------------------------


  public static boolean pngToText(String imFnm)
  /* Convert the pixels in a PNG file into a string of ASCII characters,
     and store it in a text file with the same filename + ".txt"
  */
  {
    String codeStr = pngToString(imFnm);   // convert image to an ASCII string
    if (codeStr == null)
      return false;
    else {
      String fileName = getFileName(imFnm);
      if (fileName == null)
         return false;
      else 
        return writeStringToFile(fileName+".txt", codeStr);  // save in a text file
    }
  }  // end of pngToText()



  public static String pngToString(String imFnm)
  // Convert the pixels in a PNG file into a string of ASCII chars
  {
    BufferedImage im = readImage(imFnm);    // load the image
    if (im == null) {
      System.out.println("No input image");
      return null;
    }
    else
      return convertImage(im);   // convert the image to text
  }  // end of pngToString()



  private static BufferedImage readImage(String imFnm)
  // read the image from the imFnm file
  {
    BufferedImage im = null;
    try {
      im = ImageIO.read( new File(imFnm) );
      System.out.println("Read " + imFnm);
    } 
    catch (IOException e) 
    { System.out.println("Could not read " + imFnm);  }

    return im;
  }  // end of readImage()



  private static String convertImage(BufferedImage im)
  /* Convert the pixels in the BufferedImage into a string of ASCII chars.
     Each pixel has its R, G, and B colors extracted, divided by BRIGHTNESS
     and changed to a char. The alpha component of a pixel is not examined.

    If the R, G, or B value is 0 then it is a filler value added when the
    image was created, and is skipped.

    Non-printable ASCII characters aren't stored in the string, and a count 
    of these is reported. 
  */
  {
    int numUnPrintables = 0;   // number of unprintable chars found

    StringBuilder sb = new StringBuilder();

    // extract pixels from the image into an array
    int imWidth = im.getWidth();
    int imHeight = im.getHeight();
    int[] pixels = new int[imWidth * imHeight];
    im.getRGB(0, 0, imWidth, imHeight, pixels, 0, imWidth);

    int redVal, greenVal, blueVal;
    for(int i=0; i < pixels.length; i++) {   // get the RGB values from each pixel
      // alphaVal = (pixels[i]>>24)& 255;
      redVal = (pixels[i]>>16)&255;
      greenVal = (pixels[i]>>8)&255;
      blueVal = pixels[i]&255;

      // convert the RGB values to chars, and add them to sb
      numUnPrintables += writeChar(sb, redVal);   
      numUnPrintables += writeChar(sb, greenVal);
      numUnPrintables += writeChar(sb, blueVal);
    }

    if (numUnPrintables > 0)
      System.out.println(numUnPrintables + 
               " unprintable characters were converted to spaces");

    return sb.toString();
  }  // end of convertImage()



  private static int writeChar(StringBuilder sb, int val)
  /* Write val into sb as a char and report if it's unprintable, 
     in which case a space is added to sb. Do not store 
     filler values (0's) */
  {
    if (val == 0)  // don't store a filler value
      return 0;

    int ch = val/BRIGHTNESS;  // reduce colour brightness

    if (isPrintable(ch)) {
      sb.append( (char)ch ); 
      return 0;    // char is printable
    }
    else {   // save non-ASCII char as a space
      sb.append(' '); 
      return 1;    // report that original char was unprintable
    }
  }  // end of writeChar()



  // ---------------------- file related methods -------------------------------


  private static String getFileName(String fnm)
  // extract the name from the filename by removing its suffix
  {
    int extPosn = fnm.lastIndexOf('.');
    if (extPosn == -1) {
      System.out.println("No extension found for " + fnm);
      return fnm;   // use the original file name
    }

    return fnm.substring(0, extPosn);
  }  // end of getFileName()



  private static boolean writeImageToFile(String outFnm, BufferedImage im)
  // save the im image in a PNG file called outFnm
  {
    if (!canOverWrite(outFnm))
      return false;

    boolean hasWritten = false;
    try {
      ImageIO.write(im, "png", new File(outFnm));
      System.out.println("Image written to PNG file: " + outFnm);
      hasWritten = true;
    } 
    catch(IOException e)
    { System.out.println("Could not write image to " + outFnm); }

    return hasWritten;
  } // end of writeImageToFile();



  private static boolean writeStringToFile(String outFnm, String codeStr)
  // save the codeStr string in the outFnm text file
  {
    if (!canOverWrite(outFnm))
      return false;

    boolean hasWritten = false;
    try {
      FileWriter out = new FileWriter( new File(outFnm) );
      out.write(codeStr);
      out.close();
      System.out.println("Code written to " + outFnm);
      hasWritten = true;
    }
    catch(IOException e)
    {  System.out.println("Could not write code to " + outFnm);  }

    return hasWritten;
  }  // end of writeStringToFile()



  private static boolean canOverWrite(String fnm)
  /* If fnm already exists, get a response from the
     user about whether it should be overwritten or not. */
  {
    File f = new File(fnm);
    if (!f.exists())
      return true;     // can overewrite since the file is new

    // prompt the user about whether the file can be overwritten 
    Scanner in = new Scanner(System.in);
    String response;
    System.out.print("File " + fnm + " already exists. ");
    while (true) {
      System.out.print("Overwrite (y|n)? ");
      response = in.nextLine().trim().toLowerCase();
      if (response.startsWith("n"))  // no
        return false;
      else if (response.startsWith("y"))  // yes
        return true;
    }
  }  // end of canOverWrite()


}  // end of ImageText class